// Dear emacs, this is -*- c++ -*-
// $Id$
#ifndef PFRAME_CYCLEBASE_H
#define PFRAME_CYCLEBASE_H

// STL include(s):
#include <vector>
#include <string>
#include <list>

// ROOT include(s):
#include <TSelector.h>
#include <TTree.h>
#include <TBranch.h>

namespace pf {

   /**
    * @short Base class for analysis "cycles"
    *
    *        The idea is the same as with SFrame. The user has to implement
    *        his/her analysis in a class that inherits from a common base
    *        class.
    *
    *        This class can hardly do anything at the moment, but the idea
    *        is to have something similar to SCycleBase in the end.
    */
   class CycleBase : public TSelector {

   public:
      //
      // Standard constructor and destructor:
      //
      CycleBase();
      ~CycleBase();

      //
      // Functions inherited from TSelector:
      //
      virtual Int_t  Version() const { return 2; }
      virtual void   Begin( TTree* tree );
      virtual void   SlaveBegin( TTree* tree );
      virtual void   Init( TTree* tree );
      virtual Bool_t Notify();
      virtual Bool_t Process( Long64_t entry );
      virtual void   SlaveTerminate();
      virtual void   Terminate();

   protected:
      //
      // Functions that have to be implemented by the children:
      //
      virtual void BeginInputData() = 0;
      virtual void BeginInputFile() = 0;
      virtual void ExecuteEvent( Double_t weight ) = 0;

      //
      // Some prototype convenience functions:
      //
      template< typename T >
      void ConnectVariable( const char* treeName, const char* branchName,
                            T& variable );
      template< class T >
      TBranch* DeclareVariable( T& obj, const char* name,
                                const char* treeName = 0 );
      template< class T >
      T* Book( const T& histo );

   private:
      //
      // Internal book-keeping functions and objects:
      //
      TTree* GetInputTree( const std::string& name );
      TTree* GetOutputTree( const std::string& name );
      void RegisterInputBranch( TBranch* br );
      static const char* RootType( const char* typeid_type );

      std::vector< TTree* > m_inputTrees;
      std::vector< TBranch* > m_inputBranches;

      std::vector< TTree* > m_outputTrees;
      std::list< void* > m_outputVarPointers;

      ClassDef( pf::CycleBase, 0 );

   }; // class CycleBase

} // namespace pf

#ifndef __CINT__
#include "CycleBase.icc"
#endif

#endif // PFRAME_CYCLEBASE_H
