// $Id$

// STL include(s):
#include <iostream>

// ROOT include(s):
#include <TFile.h>
#include <TTree.h>
#include <TROOT.h>

// Local include(s):
#include "../include/CycleBase.h"

ClassImp( pf::CycleBase );

namespace pf {

   CycleBase::CycleBase()
      : TSelector() {

   }

   CycleBase::~CycleBase() {

   }

   void CycleBase::Begin( TTree* /*tree*/ ) {

      return;

   }

   void CycleBase::SlaveBegin( TTree* /*tree*/ ) {

      //
      // Create the needed output TTrees:
      //
      //      const Int_t branchStyle = 1;
      //      const Int_t autoSave = 10000000;

      gROOT->cd();
      TTree* tree = new TTree( "Presel", "Format: User"
                               ", data type: MC" );

      //      tree->SetAutoSave( autoSave );
      //      TTree::SetBranchStyle( branchStyle );
      tree->SetDirectory( 0 );

      m_outputTrees.push_back( tree );
      fOutput->Add( tree );

      this->BeginInputData();

      return;

   }

   /**
    * The init function only uses the TTree parameter it receives, to
    * get the pointers to all the TTree-s in the input file. (Well, only
    * two of them for the time being.) The idea is that it doesn't matter
    * which TTree is given to this function, it should be able to access
    * all of them.
    *
    * Right now the names of the input trees are hard-coded. Later on it
    * should be a configurable property that tells the cycle which TTree-s
    * to look for.
    */
   void CycleBase::Init( TTree* tree ) {

      m_inputTrees.clear();
      m_inputBranches.clear();

      //
      // Access all the needed input TTrees:
      //
      TFile* file = tree->GetCurrentFile();
      if( ! file ) {
         std::cout << "ERROR: Couldn't get the input file pointer!"
                   << std::endl;
         return;
      }

      TTree* truth_tree = ( TTree* ) file->Get( "Truth0" );
      if( ! truth_tree ) {
         std::cout << "ERROR: Couldn't access Truth0 tree in input file!"
                   << std::endl;
         return;
      }
      m_inputTrees.push_back( truth_tree );

      TTree* reco_tree = ( TTree* ) file->Get( "FullRec0" );
      if( ! reco_tree ) {
         std::cout << "ERROR: Couldn't access FullRec0 tree in input file!"
                   << std::endl;
         return;
      }
      m_inputTrees.push_back( reco_tree );

      //
      // Call the initialisation of the child:
      //
      this->BeginInputFile();

      return;

   }

   Bool_t CycleBase::Notify() {

      return kTRUE;

   }

   /**
    * Here is where I load the current event and then call the function of
    * the child class that actually processes the event.
    */
   Bool_t CycleBase::Process( Long64_t entry ) {

      //
      // Load the current entries for all the branches:
      //
      for( std::vector< TBranch* >::const_iterator it = m_inputBranches.begin();
           it != m_inputBranches.end(); ++it ) {
         ( *it )->GetEntry( entry );
      }

      //
      // Process the event:
      //
      this->ExecuteEvent( 1.0 );

      //
      // Save the current entry in the output TTree(s):
      //
      int nbytes = 0;
      for( std::vector< TTree* >::iterator tree = m_outputTrees.begin();
           tree != m_outputTrees.end(); ++tree ) {
         nbytes = ( *tree )->Fill();
         if( nbytes < 0 ) {
            std::cout << "Tree " << ( *tree )->GetName()
                      << " write error occurred" << std::endl;
         } else if( nbytes == 0 ) {
            std::cout << "Tree " << ( *tree )->GetName()
                      << " no data written" << std::endl;
         }
      }

      return kTRUE;

   }

   void CycleBase::SlaveTerminate() {

      for( std::vector< TTree* >::iterator tree = m_outputTrees.begin();
           tree != m_outputTrees.end(); ++tree ) {
         //         ( *tree )->AutoSave();
         std::cout << "Tree: " << ( *tree )->GetName() << " has size:"
                   << ( *tree )->GetEntries() << std::endl;
      }

      return;

   }

   void CycleBase::Terminate() {

      std::cout << "In CycleBase::Terminate()" << std::endl;

      TTree* tree = dynamic_cast< TTree* >( fOutput->FindObject( "Presel" ) );

      if( tree ) {
         std::cout << "Output tree successfully accessed" << std::endl;
      } else {
         std::cout << "Output tree could not be accessed!" << std::endl;
      }

      return;

   }

   /**
    * Function returning an input TTree with a given name. Used in the
    * ConnectVariable function.
    */
   TTree* CycleBase::GetInputTree( const std::string& name ) {

      for( std::vector< TTree* >::iterator it = m_inputTrees.begin();
           it != m_inputTrees.end(); ++it ) {
         if( *it ) {
            if( name == ( *it )->GetName() ) {
               std::cout << "Found input tree with name " << name 
                         << " at " << ( *it ) << std::endl;
               return *it;
            }
         }
      }

      return 0;

   }

   TTree* CycleBase::GetOutputTree( const std::string& name ) {

      for( std::vector< TTree* >::iterator it = m_outputTrees.begin();
           it != m_outputTrees.end(); ++it ) {
         if( *it ) {
            if( name == ( *it )->GetName() ) {
               std::cout << "Found output tree with name " << name 
                         << " at " << ( *it ) << std::endl;
               return *it;
            }
         }
      }

      return 0;

   }

   /**
    * Function saving the pointer to an input branch. Used in the
    * ConnectVariable function.
    */
   void CycleBase::RegisterInputBranch( TBranch* br ) {

      if( find( m_inputBranches.begin(), m_inputBranches.end(),
                br ) != m_inputBranches.end() ) {
         std::cout << "WARNING: Branch '" << br->GetName()
                   << "' already registered!" << std::endl;
      } else {
         m_inputBranches.push_back( br );
      }

      return;

   }

   const char* CycleBase::RootType( const char* typeid_type ) {

      if( strlen( typeid_type ) != 1 ) {
         std::cout << "SCycleBaseNTuple::RootType received complex object "
                   << "description" << std::endl;
         return "";
      }

      switch( typeid_type[ 0 ] ) {

      case 'c':
         return "B";
         break;
      case 'h':
         return "b";
         break;
      case 's':
         return "S";
         break;
      case 't':
         return "s";
         break;
      case 'i':
         return "I";
         break;
      case 'j':
         return "i";
         break;
      case 'f':
         return "F";
         break;
      case 'd':
         return "D";
         break;
      case 'x':
         return "L";
         break;
      case 'y':
         return "l";
         break;
      case 'b':
         return "O";
         break;

      }

      std::cout << "Unknown primitive type encountered: "
                << typeid_type << std::endl;

      return "";

   }

} // namespace pf
